import os
import tempfile

import dash_bootstrap_components as dbc
from dash import html
from dash_extensions.enrich import Dash, FileSystemStore
from dash import dcc

from vizer.control import create_layout, create_callbacks
from vizer.ids import CHARTS_STORE_ID, SELECTED_CHART_STORE_ID, GENERAL_STORE_ID, PLOT_LAYOUT_STORE_ID, URL_ID
from vizer import workspace

STORAGE_TYPE = 'session'


def create_webapp():

    css = [dbc.themes.BOOTSTRAP] + ["https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.1/css/all.min.css"]
    fs_store = FileSystemStore(os.path.join(tempfile.gettempdir(), "file_system_store"))

    app = Dash(
        __name__, output_defaults=dict(backend=fs_store, session_check=True),
        external_stylesheets=css, prevent_initial_callbacks=True
    )

    app.layout = html.Div([
            dcc.Store(id=CHARTS_STORE_ID, data=workspace.cws.charts, storage_type=STORAGE_TYPE),
            dcc.Store(id=SELECTED_CHART_STORE_ID, data=workspace.cws.selected, storage_type=STORAGE_TYPE),
            dcc.Store(id=GENERAL_STORE_ID, data=workspace.cws.general, storage_type=STORAGE_TYPE),
            dcc.Store(id=PLOT_LAYOUT_STORE_ID, data=workspace.cws.plot, storage_type=STORAGE_TYPE),
            dcc.Location(id=URL_ID),
            create_layout()
        ], style={'height': '100vh'})

    create_callbacks(app)

    return app



import dataclasses
import json
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict
import pandas as pd


@dataclass
class Workspace:

    name: str = None
    file: str = None
    charts: Dict = field(default_factory=lambda: {'Chart1': {}})
    selected: str = 'Chart1'
    general: Dict = field(default_factory=dict)
    plot: List = field(default_factory=lambda: ['hist'] * 9)
    df: pd.DataFrame = field(default_factory=pd.DataFrame)
    path: str = None

    def save(self):
        with open(self.path, "w") as file:
            data = dataclasses.asdict(self)
            data.pop("df")
            data.pop("path")
            file.write(json.dumps(data))

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        if key == "file" and value is not None:
            self.df = pd.read_csv(value)
            for col in self.df.columns:
                if "time" in col.lower():
                    self.df[col] = pd.to_datetime(self.df[col])
            self.path = workspace_from_filename(value)


def workspace_from_filename(path):
    p = Path(os.path.abspath(path))
    r = p.parent / ("." + p.stem + ".vws")
    return str(r)


_store = {"ws": Workspace()}


def __getattr__(name):
    if name == "cws":
        return _store.get("ws")


def load_workspace(path):

    with open(path) as f:
        cws = Workspace(**json.loads(f.read()))
        cws.file = cws.file
        _store["ws"] = cws




import dash_bootstrap_components as dbc
from dash import dcc


def full_width_input(*args, **kwargs):
    if "style" in kwargs:
        kwargs["style"].update({"width": "100%"})
    else:
        kwargs["style"] = {"width": "100%"}
    return dcc.Input(*args, **kwargs)


def create_row(label, component, label_width=6):
    return dbc.Row([
        dbc.Col([
           dbc.Button(label, className="float-left", color="primary", disabled=True)
        ], width=label_width, className="mr-0 pr-0"),
        dbc.Col([
            component
        ], width=12 - label_width, className="ml-0 pl-0")
    ], className="py-1")




X_DROPDOWN_ID = "x-dropdown-id"
Y_DROPDOWN_ID = "y-dropdown-id"
Z_DROPDOWN_ID = "z-dropdown-id"
X_GRP_DROPDOWN_ID = "x-grp-dropdown-id"
Y_GRP_DROPDOWN_ID = "y-grp-dropdown-id"
Z_GRP_DROPDOWN_ID = "z-grp-dropdown-id"
GROUP_DROPDOWN_ID = "group-dropdown-id"
GROUPS_DROPDOWN_ID = "groups-dropdown-id"
AGG_DROPDOWN_ID = "agg-dropdown-id"
QUERY_INPUT_ID = "query-input-id"
ROOT_QUERY_INPUT_ID = "root-query-input-id"
OK_BUTTON_ID = "ok-button-id"
GRAPH_ID = "graph-id"
TYPE_DROPDOWN_ID = "type-dropdown-id"
CHARTS_BUTTONS_ID = "charts-buttons-id"
SELECTED_CHART_STORE_ID = "selected-chart-store-id"

CHARTS_STORE_ID = "charts-store-id"
GENERAL_STORE_ID = "general-store-id"
URL_ID = "url-id"
ADD_CHART_ID = "add-chart-id"
DELETE_CHART_ID = "delete-chart-id"
SELECT_CHECKBOX_ID = "select-checkbox-id"

PLOT_INDEX_INPUT_ID = "plot-index-input-id"
COL_INPUT_ID = "col-input-id"
PLOT_LAYOUT_STORE_ID = "plot-layout-store-id"

COL_INDEX_DROPDOWN_ID = 'col-index-dropdown-id'



from pandas import DataFrame
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype
from plotly import graph_objs as go
import numpy as np


def _create_scatter(x, y, name):
    return go.Scattergl(
        x=x, y=y, name=name,
        mode='markers', opacity=0.7,
        marker={'line': {'width': 0.5, 'color': 'white'}}
    )


AGGREGATIONS = {
    "count": lambda x: x.count(),
    "mean": lambda x: x.mean(),
    "max": lambda x: x.max(),
    "min": lambda x: x.min(),
    "std": lambda x: x.std(),
    "var": lambda x: x.var(),
    "sum": lambda x: x.sum()
}


class PlotBase:

    def __init__(self, graph_idx=None):
        self.graph_idx = graph_idx

    def _create_fig(self, x, y, z, name, idx_col=None):
        raise NotImplementedError()

    def _create_annotation(self, x, y, z):
        pass

    def generate_figure(self, df: DataFrame, query=None, x=None, y=None, z=None, group_by=None, groups=None,
                        aggregation=None, prefix='', idx_col=None, x_grp=None, y_grp=None, z_grp=None):
        df = df.copy()

        if query:
            df = df.query(query)

        if idx_col is None:
            idx_col = df.columns[0]

        chart_data = pd.concat([
            df[idx_col] if idx_col not in [x, y, x, group_by] else None,
            df[x] if x else None,
            df[y] if y else None,
            df[z] if z else None,
            df[group_by] if group_by else None,
        ], axis=1)

        # Remove duplicate columns
        chart_data = chart_data.loc[:, ~chart_data.columns.duplicated()]

        def as_string(col, value):
            if col in chart_data.select_dtypes('object').columns:
                return f"'{value}'"
            else:
                return f'{value}'

        if group_by:

            if groups:
                query = " or ".join([f"(`{group_by}` == {as_string(group_by, x)})" for x in groups])
                chart_data = chart_data.query(query)

            elif any(_k is not None for _k in [x_grp, y_grp, z_grp]):
                _x = f'{x}_{x_grp}' if x is not None else x
                _y = f'{y}_{y_grp}' if y is not None else y
                _z = f'{z}_{z_grp}' if z is not None else z

                _sub_chart = chart_data.query(f"`{group_by}` == {as_string(group_by, x_grp)}")[idx_col].to_frame()

                if x is not None:
                    _df = chart_data.query(f"`{group_by}` == {as_string(group_by, x_grp)}")[[idx_col, x]].rename(columns={x: _x})
                    _sub_chart = _sub_chart.merge(_df, how='outer', left_on=idx_col, right_on=idx_col)

                if y is not None:
                    _df = chart_data.query(f"`{group_by}` == {as_string(group_by, y_grp)}")[[idx_col, y]].rename(columns={y: _y})
                    _sub_chart = _sub_chart.merge(_df, how='outer', left_on=idx_col, right_on=idx_col)

                if z is not None:
                    _df = chart_data.query(f"`{group_by}` == {as_string(group_by, z_grp)}")[[idx_col, z]].rename(columns={z: _z})
                    _sub_chart = _sub_chart.merge(_df, how='outer', left_on=idx_col, right_on=idx_col)

                x = _x
                y = _y
                z = _z

                chart_data = _sub_chart.dropna()

        if idx_col:
            chart_data.sort_values([idx_col])

        # if aggregation:
        #     to_grp = [x]
        #     if group_by:
        #         to_grp += [group_by]
        #     chart_data = chart_data.groupby(to_grp)[[y]]
        #     chart_data = AGGREGATIONS[aggregation](chart_data).reset_index()

        chart_data = chart_data.dropna()
        if x:
            chart_data = chart_data.sort_values(by=x)

        if group_by is not None and groups:
            data = []
            annotations = []
            for grp in groups:
                sub = chart_data.query(f"`{group_by}` == {as_string(group_by, grp)}")
                data.append(self._create_fig(sub[x], sub[y], sub[z] if z else None, f"{prefix}_" + str(grp), sub[idx_col]))
                annotations.append(self._create_annotation(sub[x], sub[y], sub[z] if z else None))
        else:
            data = [self._create_fig(chart_data[x], chart_data[y], chart_data[z] if z else None, prefix,
                                     chart_data[idx_col])]
            annotations = [self._create_annotation(chart_data[x], chart_data[y], chart_data[z] if z else None)]
        return data, annotations


def r_squared(obs, exp):

    if not (is_numeric_dtype(obs) and is_numeric_dtype(exp)):
        return np.nan
    y_mean = np.mean(obs)
    ss_tot = sum((obs - y_mean)**2)
    ss_res = sum((obs - exp)**2)
    if ss_tot == 0.0:
        return np.nan

    return 1.0 - ss_res / ss_tot


def rmse(data, exp):
    if not (is_numeric_dtype(data) and is_numeric_dtype(exp)):
        return np.nan
    return np.sqrt(np.mean((data - exp)**2.0))


class Line(PlotBase):
    def _create_fig(self, x, y, z, name, idx_col=None):
        return go.Scattergl(
            x=x, y=y, name=name, hovertext=idx_col,
            mode='lines', opacity=0.7
        )

    def _create_annotation(self, x, y, z):
        r2 = r_squared(x, y)
        _rmse = rmse(x, y)
        if np.isnan(r2):
            return []
        return [f"r2={r2:0.3f}", f"rmse={_rmse:0.3f}"]


class Scatter(PlotBase):
    def _create_fig(self, x, y, z, name, idx_col=None):
        return go.Scattergl(
            x=x, y=y, name=name, hovertext=idx_col,
            mode='markers', opacity=0.7,
            marker={'line': {'width': 0.5, 'color': 'white'}}
        )

    def _create_annotation(self, x, y, z):
        r2 = r_squared(x, y)
        _rmse = rmse(x, y)
        if np.isnan(r2):
            return []
        return [f"r2={r2:0.3f}", f"rmse={_rmse:0.3f}"]


class Hist(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):
        fig = go.Histogram(x=y, name=name, nbinsx=100)
        return fig

    def _create_annotation(self, x, y, z):
        return [f"mean={np.mean(y):0.3f}", f"std={np.std(y):0.3f}"]


class Hist2D(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):
        fig = go.Histogram2d(x=x, y=y, name=name, nbinsx=100, nbinsy=100, coloraxis='coloraxis')
        return fig

    def _create_annotation(self, x, y, z):
        return [f"mean_x={np.mean(x):0.3f}", f"std_x={np.std(x):0.3f}", f"mean_y={np.mean(y):0.3f}", f"std_y={np.std(y):0.3f}"]


class HistCum(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):
        fig = go.Histogram(x=y, name=name, cumulative_enabled=True)
        return fig

    def _create_annotation(self, x, y, z):
        return [f"mean={np.mean(y):0.3f}", f"std={np.std(y):0.3f}"]


class Table(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):

        header = []
        cells = []
        for col in [idx_col, x, y, z]:
            if col is not None:
                header.append(col.name)
                cells.append(col)

        return go.Table(
           header=dict(
               values=header
           ),
           cells=dict(
               values=cells
           )
        )


class Heatmap(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):
        return go.Scatter(x=x, y=y, name=name, mode='markers', marker=dict(
            size=16,
            color=z,
            colorscale='Viridis',
            showscale=True,
            line={'width': 0.5, 'color': 'white'},
            coloraxis='coloraxis'
        ))


class Map(PlotBase):

    def _create_fig(self, x, y, z, name, idx_col=None):

        if idx_col is not None and z is not None:
            hover_text = [f"{_x}={_z}" for _x, _z in zip(idx_col, z)]
        elif z is not None:
            hover_text = [f"{_z}" for _z in z]
        else:
            hover_text = None

        return go.Scattermapbox(
            lat=x, lon=y, hovertext=hover_text, name=name,
            marker=dict(
                size=16,
                color=z,
                colorscale='Viridis',
                showscale=True,
                coloraxis='coloraxis'
            )
        )



import os
import sys

import socket
from contextlib import closing
import webbrowser

import logging

from dash import Input, Dash

from vizer.gui import create_webapp
from vizer.ids import CHARTS_STORE_ID, PLOT_LAYOUT_STORE_ID, SELECTED_CHART_STORE_ID, GENERAL_STORE_ID
from vizer import workspace

log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def register_auto_save(app: Dash):
    @app.callback(
        Input(CHARTS_STORE_ID, "data"), Input(SELECTED_CHART_STORE_ID, "data"), Input(GENERAL_STORE_ID, "data"),
        Input(PLOT_LAYOUT_STORE_ID, "data")
    )
    def save_workspace(charts, selected, general, plot):
        workspace.cws.plot = plot
        workspace.cws.general = general
        workspace.cws.selected = selected
        workspace.cws.charts = charts
        if workspace.cws.path:
            workspace.cws.save()


def run():

    if len(sys.argv) > 1:

        if sys.argv[1].endswith(".vws"):
            workspace.load_workspace(sys.argv[1])
            print("Loaded workspace")
        else:
            try:
                workspace.load_workspace(workspace.workspace_from_filename(sys.argv[1]))
                print("Found workspace - loading")
            except FileNotFoundError:
                workspace.cws.file = os.path.abspath(sys.argv[1])
                print("Loaded file in empty workspace")

    app = create_webapp()
    register_auto_save(app)

    p = find_free_port()
    webbrowser.open(f"http://localhost:{p}")
    app.run_server(port=p, host="0.0.0.0", debug=False)


if __name__ == "__main__":
    cli = sys.modules['flask.cli']
    cli.show_server_banner = lambda *x: None
    run()



import itertools
import math

import dash_bootstrap_components as dbc
from dash import dcc
from dash import html
import plotly
from dash.dependencies import ALL
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Dash, Trigger, Output, Input, State
from dash_extensions.snippets import get_triggered

import plotly.graph_objects as go
from plotly.graph_objs import Scattermapbox

from vizer import workspace
from vizer.graph import Scatter, Hist, AGGREGATIONS, HistCum, Table, Heatmap, Map, Hist2D, Line
from vizer.utils import create_row, full_width_input

from vizer.ids import *


def create_layout():
    return html.Div([
        dbc.Card([
            dbc.CardHeader([
                dbc.Row([
                    dbc.Col([
                        html.Div(id=CHARTS_BUTTONS_ID)
                    ], width=5),
                    dbc.Col([
                        dbc.Button(html.I(className="fa fa-trash"), id=DELETE_CHART_ID, color="danger", className="mx-1 float-right"),
                        dbc.Button(html.I(className="fa fa-plus"), id=ADD_CHART_ID, color="success", className="mx-1 float-right")
                    ], width=1),
                    dbc.Col([
                        create_row("Index column", dcc.Dropdown(id=COL_INDEX_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Root query", full_width_input(id=ROOT_QUERY_INPUT_ID, type="text"), 3)
                    ], width=4)
                ])
            ]),
            dbc.CardBody([
                dbc.Row([
                    dbc.Col([
                        create_row("Plot Index", dcc.Dropdown(
                            id=PLOT_INDEX_INPUT_ID, options=[{'value': x, 'label': str(x)} for x in range(1,10)]), 6)
                    ], width=2),
                    dbc.Col([
                        create_row("X", dcc.Dropdown(id=X_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Y", dcc.Dropdown(id=Y_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Z", dcc.Dropdown(id=Z_DROPDOWN_ID))
                    ], width=2)
                ]),
                dbc.Row([
                    dbc.Col([
                        create_row("Group by", dcc.Dropdown(id=GROUP_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("X group", dcc.Dropdown(id=X_GRP_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Y group", dcc.Dropdown(id=Y_GRP_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Z group", dcc.Dropdown(id=Z_GRP_DROPDOWN_ID))
                    ], width=2),
                    dbc.Col([
                        create_row("Groups", dcc.Dropdown(id=GROUPS_DROPDOWN_ID, multi=True), 2)
                    ], width=4)
                ]),
                dbc.Row([
                    dbc.Col([
                        create_row("Aggregation", dcc.Dropdown(id=AGG_DROPDOWN_ID,
                                                               options=[{'value': x, 'label': x} for x in
                                                                        AGGREGATIONS.keys()], disabled=True))
                    ], width=2),
                    dbc.Col([
                        create_row("Chart query", full_width_input(id=QUERY_INPUT_ID, type="text"), 2)
                    ], width=6),
                    dbc.Col([
                        dcc.Checklist(
                            id=SELECT_CHECKBOX_ID,
                            options=[
                                {'label': 'Select', 'value': 'select'}
                            ],
                            value=[]
                        )
                    ], width=4)
                ])
            ])
        ], className="m-1"),
        html.Div([
            html.Div(id=GRAPH_ID, style={"height": "100%"})
        ], className="flex-grow-1 m-1")

    ], style={"height": "100%", "width": "100%", "background-color": "#D3D3D3"}, className="d-flex h-100 flex-column")


def create_callbacks(app: Dash):

    @app.callback(
        Trigger(URL_ID, 'pathname'),
        Output(X_DROPDOWN_ID, "options"), Output(Y_DROPDOWN_ID, "options"), Output(Z_DROPDOWN_ID, "options"),
        Output(X_GRP_DROPDOWN_ID, "options"), Output(Y_GRP_DROPDOWN_ID, "options"), Output(Z_GRP_DROPDOWN_ID, "options"),
        Output(GROUPS_DROPDOWN_ID, "options"), Output(GROUP_DROPDOWN_ID, "options"), Output(COL_INDEX_DROPDOWN_ID, "options"),
        Output(COL_INDEX_DROPDOWN_ID, "value"),
        State(CHARTS_STORE_ID, 'data'), State(SELECTED_CHART_STORE_ID, 'data'), State(GENERAL_STORE_ID, 'data')
    )
    def update_options(charts, selected_chart, general):

        _df = workspace.cws.df

        group_by = charts[selected_chart].get("args", {}).get("group_by")
        if group_by:
            group_by_cols = [{'value': str(x), 'label': str(x)} for x in _df[str(group_by)].unique()]
        else:
            group_by_cols = []

        cols = [{'value': x, 'label': x} for x in _df.columns]

        return cols, cols, cols, group_by_cols, group_by_cols, group_by_cols, group_by_cols, cols, cols, general.get("col_idx")

    @app.callback(
        Trigger(dict(type="chart-selector", chart=ALL), "n_clicks"),
        Output(SELECTED_CHART_STORE_ID, 'data')
    )
    def on_selection():
        t = get_triggered()
        if t.n_clicks is None:
            raise PreventUpdate()
        return t.id["chart"]

    @app.callback(
        [Trigger(URL_ID, 'pathname'), Trigger(ADD_CHART_ID, 'n_clicks'), Trigger(DELETE_CHART_ID, 'n_clicks'),
         Input(SELECTED_CHART_STORE_ID, 'data'), Input(CHARTS_STORE_ID, 'data')],
        [Output(CHARTS_BUTTONS_ID, "children")]
    )
    def update_charts_buttons(selected, charts):

        def color(c):
            if c == selected:
                return "primary"
            else:
                return "secondary"

        k = [dbc.Button(x, color=color(x), id=dict(type="chart-selector", chart=x), className="mx-1") for x in charts.keys()]

        return k

    @app.callback(
        [Trigger(URL_ID, 'pathname'), Trigger(ADD_CHART_ID, 'n_clicks'), Trigger(DELETE_CHART_ID, 'n_clicks')],
        [Output(CHARTS_STORE_ID, 'data'), Output(ROOT_QUERY_INPUT_ID, 'value')],
        [State(CHARTS_STORE_ID, 'data'), State(GENERAL_STORE_ID, 'data'), State(SELECTED_CHART_STORE_ID, 'data')], group="charts"
    )
    def update_or_add_charts_dropdown(charts, general, current):

        trig = get_triggered()

        if trig.id == ADD_CHART_ID:
            last = sorted(charts.keys())[-1]
            value = f"Chart{int(last.lstrip('Chart')) + 1}"
            charts[value] = charts[last].copy()
            if "index" in charts[value]:
                charts[value]["index"] += 1
        elif trig.id == DELETE_CHART_ID:
            del charts[current]

        if not charts:
            charts = {'Chart1': {}}

        return charts, general.get("root_query")

    @app.callback(
        [Trigger(URL_ID, 'pathname'), Input(SELECTED_CHART_STORE_ID, 'data'), Input(GROUP_DROPDOWN_ID, 'value')],
        [Output(X_DROPDOWN_ID, 'value'), Output(Y_DROPDOWN_ID, 'value'), Output(Z_DROPDOWN_ID, 'value'),
         Output(GROUP_DROPDOWN_ID, 'value'), Output(GROUPS_DROPDOWN_ID, 'value'), Output(QUERY_INPUT_ID, 'value'), Output(AGG_DROPDOWN_ID, 'value'),
         Output(GROUPS_DROPDOWN_ID, 'options'), Output(PLOT_INDEX_INPUT_ID, 'value'),
         Output(X_GRP_DROPDOWN_ID, 'options'), Output(Y_GRP_DROPDOWN_ID, 'options'), Output(Z_GRP_DROPDOWN_ID, 'options'),
         Output(X_GRP_DROPDOWN_ID, 'value'), Output(Y_GRP_DROPDOWN_ID, 'value'), Output(Z_GRP_DROPDOWN_ID, 'value')],
        State(CHARTS_STORE_ID, 'data'), group="inputs"
    )
    def update_chart_on_selection(selected_chart, group, charts):

        df = workspace.cws.df

        d = charts[selected_chart]
        kwargs = charts[selected_chart].get("args", {})

        if get_triggered().id == GROUP_DROPDOWN_ID:
            if group:
                _group = group
                _selected_groups = df[group].unique()[:10]
            else:
                _group = None
                _selected_groups = None
        else:
            _group = kwargs.get("group_by")
            _selected_groups = kwargs.get("groups")

        groups_options = [{'value': x, 'label': x} for x in df[_group].unique()] if _group is not None else []
        return kwargs.get("x"), kwargs.get("y"), kwargs.get("z"), _group, \
               _selected_groups, kwargs.get("query"), kwargs.get("aggregation"), groups_options, d.get('index', 1), \
               groups_options, groups_options, groups_options, kwargs.get("x_grp"), kwargs.get("y_grp"), kwargs.get("z_grp")

    @app.callback(
        Input(dict(type=TYPE_DROPDOWN_ID, index=ALL), 'value'),
        Output(PLOT_LAYOUT_STORE_ID, 'data'),
        State(PLOT_LAYOUT_STORE_ID, 'data')
    )
    def on_plot_type_change(new_type, layouts):
        idx = get_triggered().id['index'] - 1
        layouts[idx] = new_type[idx]
        return layouts

    @app.callback(
        Input(ROOT_QUERY_INPUT_ID, 'value'), Input(COL_INDEX_DROPDOWN_ID, 'value'),
        Output(GENERAL_STORE_ID, 'data'),
        State(GENERAL_STORE_ID, 'data')
    )
    def on_general_query(val, col_idx, general):
        general["root_query"] = val
        general["col_idx"] = col_idx
        return general

    @app.callback(
        [Output("query-modal", "is_open"), Output("modal-content", "children")],
        [Input(dict(type="modal-open", idx=ALL), "n_clicks"), Input("close", "n_clicks")],
        [State("query-modal", "is_open"), State(CHARTS_STORE_ID, 'data')],
    )
    def toggle_modal(n1, n2, is_open, charts):
        t = get_triggered()
        cnt = []
        if isinstance(t.id, dict):
            plot_idx = t.id['idx']

            for name, val in charts.items():
                if val['index'] == plot_idx:
                    cnt.append(dbc.Row(f"{name}: {val['args']['query']}"))

        if n1 or n2:
            return not is_open, cnt
        return is_open, cnt

    @app.callback(
        [Input(CHARTS_STORE_ID, 'data'), Input(PLOT_LAYOUT_STORE_ID, 'data'), Input(GENERAL_STORE_ID, 'data')],
        Output(GRAPH_ID, 'children')
    )
    def do_plot(charts, plot_types, general):
        df = workspace.cws.df
        try:
            figures = [go.Figure(layout=dict(margin=dict(t=5, l=10, r=10, b=10))) for _ in range(9)]
            graphs = [html.Div() for _ in range(9)]
            annotations = [[] for _ in range(9)]

            for chart_name, chart_data in charts.items():
                if not chart_data:
                    continue
                _p_index = chart_data.get('index', 0)

                plot_type = plot_types[_p_index - 1]

                kwargs = chart_data.get("args", {}).copy()
                kwargs["prefix"] = chart_name

                queries = []
                if general.get("root_query"):
                    queries.append(general.get("root_query"))
                if kwargs.get("query"):
                    queries.append(kwargs.get("query"))
                if general.get("col_idx"):
                    kwargs["idx_col"] = general.get("col_idx")

                if general.get("root_query"):
                    kwargs["query"] = ' & '.join([f'({x})' for x in queries])


                fig = figures[_p_index - 1]

                if plot_type == "cumhist":
                    p = HistCum()
                    title = kwargs['y']
                    x_title = kwargs['y']
                    y_title = "Counts"
                elif plot_type == "hist":
                    p = Hist()
                    title = kwargs['y']
                    x_title = kwargs['y']
                    y_title = "Counts"
                    fig.update_layout(barmode='overlay')
                    fig.update_traces(opacity=0.50)
                elif plot_type == "hist2d":
                    p = Hist2D()
                    title = kwargs['y']
                    x_title = kwargs['x']
                    y_title = kwargs['y']
                elif plot_type == "table":
                    p = Table()
                    title = kwargs['y']
                    x_title = kwargs['x']
                    y_title = kwargs['y']
                elif plot_type == "heatmap":
                    p = Heatmap()
                    title = kwargs['z']
                    x_title = kwargs['x']
                    y_title = kwargs['y']
                elif plot_type == "map":
                    p = Map()
                    title = kwargs['z']
                    x_title = kwargs['x']
                    y_title = kwargs['y']
                    fig.update_layout(mapbox_style="open-street-map")
                elif plot_type == "line":
                    p = Line()
                    title = f"{kwargs['x']} vs. {kwargs['y']}"
                    x_title = kwargs['x']
                    y_title = kwargs['y']
                else:
                    p = Scatter()
                    title = f"{kwargs['x']} vs. {kwargs['y']}"
                    x_title = kwargs['x']
                    y_title = kwargs['y']

                figs, _annotations = p.generate_figure(df.copy(), **kwargs)

                for k in figs:
                    fig.add_trace(k)

                for y in _annotations:
                    if y:
                        annotations[_p_index - 1].append('<br>'.join([kwargs['y']] + y))

                fig.update_layout(dict(uirevision="JOHN"))
                fig.update_layout(dict(xaxis=dict(title=x_title)))
                fig.update_layout(dict(yaxis=dict(title=y_title)))
                fig.update_layout(legend=dict(
                    orientation="h",
                    yanchor="bottom",
                    y=1.02,
                    xanchor="left",
                    x=0
                ))

                graphs[_p_index - 1] = html.Div([
                    dbc.Card([
                        dbc.CardHeader([
                            dbc.Row([
                                dbc.Col([
                                    html.Div([f'{title}']),

                                    dbc.Modal(
                                        [
                                            dbc.ModalHeader("Queries"),
                                            dbc.ModalBody("This is the content of the modal", id='modal-content'),
                                            dbc.ModalFooter(
                                                dbc.Button("Close", id="close", className="ml-auto")
                                            ),
                                        ],
                                        id="query-modal",
                                    )],
                                    width=6
                                ),
                                dbc.Col(
                                    dbc.Button(
                                        "Query", id=dict(type="modal-open", idx=_p_index), color="primary"
                                    ) if chart_data["args"]["query"] else None,
                                    width=2
                                ),
                                dbc.Col(
                                    html.Div(
                                        dcc.Dropdown(
                                            id=dict(type=TYPE_DROPDOWN_ID, index=_p_index),
                                            options=[
                                                {'value': 'hist', 'label': 'Histogram'},
                                                {'value': 'hist2d', 'label': 'Histogram2D'},
                                                {'value': 'cumhist', 'label': 'CumHistogram'},
                                                {'value': 'scat', 'label': 'Scatter'},
                                                {'value': 'line', 'label': 'Line'},
                                                {'value': 'table', 'label': 'Table'},
                                                {'value': 'map', 'label': 'Map'},
                                                {'value': 'heatmap', 'label': 'Heatmap'}
                                            ],
                                            value=plot_type,
                                            clearable=False,
                                            style={"width": "120px"}
                                        ),
                                        className="float-right float-end"
                                    ),
                                    width=4
                                ),
                            ], justify="between"),
                        ]),
                        dbc.CardBody(
                            dcc.Graph(figure=fig, style={"height": f"100%"}, id=dict(type="graph", index=_p_index))
                        )
                    ], style={"height": f"100%"})
                ], className="m-1")

            for fig, a in zip(figures, annotations):
                if fig.data and isinstance(fig.data[0], Scattermapbox):
                    try:
                        lat = [list(x.lat) for x in fig.data]
                        lon = [list(x.lon) for x in fig.data]

                        lat = list(itertools.chain(*lat))
                        lon = list(itertools.chain(*lon))

                        diff_lat = (max(lat) - min(lat))
                        diff_lon = (max(lon) - min(lon))

                        center_lat = diff_lat / 2.0 + min(lat)
                        center_lon = diff_lon / 2.0 + min(lon)

                        fig.update_layout(mapbox=dict(
                            center=go.layout.mapbox.Center(
                                lat=center_lat,
                                lon=center_lon
                            ),
                            zoom=5.5 - len(fig.data)
                        ))
                    except:
                        pass

                if any(a):
                    i = 0
                    offset = 0
                    for label in a:

                        fig.add_annotation(
                            x=0 + offset,
                            y=1.0,
                            xref="paper",
                            yref="paper",
                            text=label,
                            showarrow=False,
                            font=dict(
                                family="Courier New, monospace",
                                size=12,
                                color="black"
                            ),
                            align="right",
                            bgcolor=plotly.colors.qualitative.Plotly[i],
                            opacity=0.8
                        )
                        i += 1

                        offset += 0.2

            graphs = [g for g, f in zip(graphs, figures) if len(f.data)]

            rows = math.ceil(len(graphs) / 3)
            cols = min(len(graphs), 3)

            columns = " ".join([f"{100 / cols}%"] * cols) if cols else "100%"
            rows = " ".join([f"{100 / rows}%"] * rows) if rows else "100%"

            return html.Div(graphs, style={"height": "99%", "display": "grid", "grid-template-columns": columns, "grid-template-rows": rows}),

        except Exception as e:
            #import traceback
            #traceback.print_exc()
            raise PreventUpdate()

    @app.callback(
        Input(dict(type="graph", index=ALL), 'selectedData'),
        Output(QUERY_INPUT_ID, 'value'),
        [State(dict(type="graph", index=ALL), 'figure'), State(SELECTED_CHART_STORE_ID, 'data')], group="inputs"
    )
    def on_box_select(d, layout, do_select):

        if not do_select:
            raise PreventUpdate()

        t = get_triggered()
        idx = t.id['index'] - 1

        x_label = layout[idx]['layout']['xaxis']['title']['text']
        y_label = layout[idx]['layout']['yaxis']['title']['text']

        try:
            x_range = d[0]["range"]['x']
            y_range = d[0]["range"]['y']

            sub_queries = []
            sub_queries.append(f"{x_range[0]:0.4f} < {x_label}")
            sub_queries.append(f"{x_label} < {x_range[1]:0.4f}")

            sub_queries.append(f"{y_range[0]:0.4f} < {y_label}")
            sub_queries.append(f"{y_label} < {y_range[1]:0.4f}")

            query = ' & '.join(sub_queries)
            return query

        except (KeyError, TypeError):
            raise PreventUpdate()

    @app.callback(
        Input(dict(type="graph", index=ALL), 'relayoutData'),
        [Output(QUERY_INPUT_ID, 'value')],
        [State(SELECTED_CHART_STORE_ID, 'data'), State(CHARTS_STORE_ID, 'data'),
         State(dict(type="graph", index=ALL), 'figure'), State(SELECT_CHECKBOX_ID, 'value')], group="inputs"
    )
    def doo(d, current, charts, layout, do_select):
        if not do_select:
            raise PreventUpdate()

        df = workspace.cws.df

        try:
            t = get_triggered()
            idx = t.id['index'] - 1

            x_label = layout[idx]['layout']['xaxis']['title']['text']
            y_label = layout[idx]['layout']['yaxis']['title']['text']
            g = d[idx]
            sub_queries = []

            query = charts[current]['args']['query']

            if 'xaxis.range[0]' in d or 'yaxis.range[0]' in g:
                if x_label in df.columns:
                    try:
                        sub_queries.append(f"{g['xaxis.range[0]']:0.4f} < {x_label}")
                        sub_queries.append(f"{x_label} < {g['xaxis.range[1]']:0.4f}")
                    except (StopIteration, KeyError):
                        pass

                if y_label in df.columns:
                    try:
                        sub_queries.append(f"{g['yaxis.range[0]']:0.4f} < {y_label}")
                        sub_queries.append(f"{y_label} < {g['yaxis.range[1]']:0.4f}")
                    except (StopIteration, KeyError):
                        pass

                query = ' & '.join(sub_queries)
            elif g == {'xaxis.autorange': True, 'yaxis.autorange': True}:
                query = ''
            return query
        except Exception as e:
            #import traceback
            #traceback.print_exc()
            raise PreventUpdate()

    @app.callback(
        [Input(X_DROPDOWN_ID, 'value'),
         Input(Y_DROPDOWN_ID, 'value'), Input(Z_DROPDOWN_ID, 'value'), Input(GROUP_DROPDOWN_ID, 'value'), Input(GROUPS_DROPDOWN_ID, 'value'),
         Input(QUERY_INPUT_ID, 'value'), Input(AGG_DROPDOWN_ID, 'value'), Input(SELECTED_CHART_STORE_ID, 'data'),
         Input(PLOT_INDEX_INPUT_ID, 'value'),
         Input(X_GRP_DROPDOWN_ID, 'value'), Input(Y_GRP_DROPDOWN_ID, 'value'), Input(Z_GRP_DROPDOWN_ID, 'value')],
        [Output(CHARTS_STORE_ID, 'data')],
        [State(CHARTS_STORE_ID, 'data')], group="charts"
    )
    def ok(x, y, z, group, groups, query, agg, selected_chart, plot_index, x_grp, y_grp, z_grp, charts):
        kwargs = dict(x=x, y=y, z=z, group_by=group, groups=groups, query=query, aggregation=agg,
                      x_grp=x_grp, y_grp=y_grp, z_grp=z_grp)
        charts[selected_chart] = {"args": kwargs, "index": plot_index}
        return charts



#!/usr/bin/env python

from setuptools import setup
import pathlib
import os

HERE = pathlib.Path(__file__).parent
long_description = (HERE / "README.md").read_text()

setup(
    name='vizu',
    version=os.getenv("VERSION_TAG", "v0.0.2").lstrip("v"),
    description='Visualization of data files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jesper Halkjær Jensen',
    author_email='mail@jeshj.com',
    url='https://github.com/gedemagt/vizer',
    packages=['vizer', 'vizer.assets'],
    entry_points={
        'console_scripts': [
            'vizu = vizer.main:run',
        ],
    },
    license="MIT",
    python_requires='>=3.6',
    install_requires=[
        "flask==2.0.3",
        "pandas==1.4.1",
        "scipy==1.8.0",
        "dash==2.2.0",
        "dash-extensions==0.0.71",
        "dash-bootstrap-components==1.0.3",
        "dash-table==5.0.0"
    ]
)
